"""
Compatibility layer and utilities.
"""
import sys

PY3 = sys.version_info[0] >= 3

if PY3:
    from io import StringIO  # noqa
    import urllib.parse as urlparse

    def iteritems(d):
        return iter(d.items())

    def itervalues(d):
        return iter(d.values())

    str_types = (str)
else:
    import urlparse
    from StringIO import StringIO  # noqa

    def iteritems(d):
        return d.iteritems()

    def itervalues(d):
        return d.itervalues()

    str_types = (str, unicode)



from __future__ import absolute_import

import os
from elasticsearch import RequestsHttpConnection
from requests_aws4auth import AWS4Auth
from elasticsearch_dsl.connections import connections


class FlaskElasticsearch(object):

    EXTENSION_NAME = 'flask_elasticsearch'

    def __init__(self, app=None, **kwargs):
        if app is not None:
            self.init_app(app, **kwargs)

    def init_app(self, app, **kwargs):

        if not hasattr(app, 'extensions'):
            app.extensions = {}

        try:
            options = {
                'hosts': app.config['ES_ENDPOINT'],
                'http_auth': app.config.get('ES_HTTP_AUTH', None),
                'timeout': app.config.get('ES_TIMEOUT', 10),
                'use_ssl': app.config.get('ES_USE_SSL', True),
                'verify_certs': app.config.get('ES_VERIFY_CERTS', True),
                'ca_certs': app.config.get('ES_CA_CERTS', None),
                'client_cert': app.config.get('ES_CLIENT_CERT', None),
                'client_key': app.config.get('ES_CLIENT_KEY', None),
                'headers': app.config.get('ES_HEADERS', None),
                'aws_access_key_id': app.config.get('ES_AWS_ACCESS_KEY', None),
                'aws_secret_access_key':
                    app.config.get('ES_AWS_SECRET_KEY', None),
                'aws_region': app.config.get('ES_AWS_REGION', 'us-east-1')
            }
            options.update(**kwargs)

            self.client = ESClient(**options).create_connection()
        except Exception as e:
            self.client = None

        app.extensions[self.EXTENSION_NAME] = self
        self.app = app


class ESClient(object):
    """ Return an elasticsearch client.

    Arguments:
        hosts: Expects an array of objects with host information for
               the elasticsearch cluster connection. In form:
               [{'host': es_endpoint, 'port': 443}]
        http_auth: Authentication details. If None, this will attempt to create
                   an AWS4Auth auth object, which requires aws access/secret
        use_ssl: Boolean
        verify_certs: Boolean
        verify_certs: Boolean
        ca_certs: Path to CA bundle. By default standard requests' bundle will
                  be used.
        client_cert: Path to the file containing the private key and the
                     certificate, or cert only if using client_key
        client_key: Path to the file containing the private key if using
                    separate cert and key files (client_cert will contain only
                    the cert)
        headers: Any custom http headers to be add to requests
        aws_access_key_id: Access Key for authorized IAM user.
        aws_secret_access_key: Secret Key for authorized IAM user.
        aws_region: Region for connection (e.g. us-east-1)

    Currently only supports the RequestsHttpConnection Transport.
    http://elasticsearch-py.readthedocs.io/en/master/transports.html
    """
    def __init__(self, hosts=['localhost'], http_auth=None, timeout=10,
                 use_ssl=True, verify_certs=True, ca_certs=None,
                 client_cert=None, client_key=None, headers=None,
                 aws_access_key_id=None, aws_secret_access_key=None,
                 aws_region='us-east-1', **kwargs):
        super(ESClient, self).__init__()

        self.hosts = hosts
        self.http_auth = http_auth
        self.timeout = timeout
        self.use_ssl = use_ssl
        self.verify_certs = verify_certs
        self.ca_certs = ca_certs
        self.client_cert = client_cert
        self.client_key = client_key
        self.headers = headers
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_region = aws_region

        self.auto_http_auth_kinds = ('aws4auth')

    def _get_http_auth(self):
        """ Convenience method to get various kinds of HTTP auth objects.

            Currently supports creating an AWS4Auth auth object.
        """
        if self.http_auth is None:
            return None

        # If this isn't a string, pass it on as-is. Error will be thrown when
        # trying to establish connection if it isnt valid.
        if not isinstance(self.http_auth, str):
            return self.http_auth

        # If it is a string but doesn't exist in our list of auth kinds we
        # can create automatically, return it unchanged.
        elif self.http_auth not in self.auto_http_auth_kinds:
            return self.http_auth

        # We should know how to create a connection object.
        elif self.http_auth in self.auto_http_auth_kinds:
            if self.http_auth.lower() == 'aws4auth'\
                    and self.aws_access_key_id is not None\
                    and self.aws_secret_access_key is not None\
                    and self.aws_region is not None:
                return AWS4Auth(
                    self.aws_access_key_id,
                    self.aws_secret_access_key,
                    self.aws_region,
                    'es'
                )

        return None

    @classmethod
    def from_env(cls, hosts_var='ES_ENDPOINT', port=443,
                 timeout_var='ES_TIMEOUT', use_ssl_var='ES_USE_SSL',
                 verify_certs_var='ES_VERIFY_CERTS', ca_certs_var='ES_CA_CERTS',
                 client_cert_var='ES_CLIENT_CERT',
                 client_key_var='ES_CLIENT_KEY', headers_var='ES_HEADERS',
                 aws_access_key_id_var='ES_AWS_ACCESS_KEY',
                 aws_secret_access_key_var='ES_AWS_SECRET_KEY',
                 aws_region_var='ES_AWS_REGION', **kwargs):
        """ Alternate constructor for building an ESClient object from
        environment variables.  Particularly useful in combination with
        honcho/.env files.

        Each *_var argument corresponds to an argument in init.  If the
        variable is not in the enviroment, it will look in *args, **kwargs,
        and then finally fall back to the defaults in init if not present
        elsewhere.

        Typically this should be called as
        client = ESClient.from_env(http_ath='aws4auth')

        Args:
            hosts_var (str): Name of the environment variable corresponding
                to self.hosts
            port (str): Name of the environment variable corresponding
                to self.port
            timeout_var (str): Name of the environment variable corresponding
                to self.timeout
            use_ssl_var (str): Name of the environment variable corresponding
                to self.use_ssl
            verify_certs_var (str): Name of the environment variable
                corresponding to self.verify_certs
            ca_certs_var (str): Name of the environment variable corresponding
                to self.ca_certs
            client_cert_var (str): Name of the environment variable
                corresponding to self.client_cert
            client_key_var (str): Name of the environment variable corresponding
                to self.client_key
            headers_var (str): Name of the environment variable corresponding
                to self.headers
            aws_access_key_id_var (str): Name of the environment variable
                corresponding to self.aws_access_key_id
            aws_secret_access_key_var (str): Name of the environment variable
                corresponding to self.aws_secret_access_key
            aws_region_var (str): Name of the environment variable corresponding
                to self.aws_region

        Returns: Initialized class instance.

        """

        def check_vars(var_name, kwarg_name, result_dict):
            """ Check if a variable is in the environment, if so update the
                dict of results to send to init.

            Args:
                var_name: e.g., 'ES_ENDPOINT'
                kwarg_name: the name of the kwarg expected by init.  E.g.
                    'hosts'
                result_dict (dict): a dict mapping kwargs to values in init

            Returns (dict): updated result_dict

            """
            var_val = os.environ.get(var_name, None)
            if var_val is not None:
                result_dict[kwarg_name] = var_val
            return result_dict

        result_dict = {}
        result_dict = check_vars(hosts_var, 'hosts', result_dict)
        if 'hosts' in result_dict:
            result_dict['hosts'] = [{'host': result_dict['hosts'],
                                     'port': port}]

        kwarg_names = ['timeout', 'use_ssl', 'verify_certs', 'ca_certs',
                       'client_cert', 'client_key', 'headers',
                       'aws_access_key_id', 'aws_secret_access_key',
                       'aws_region']
        var_names = [x + '_var' for x in kwarg_names]
        for var, kwarg_name in zip(var_names, kwarg_names):
            result_dict = check_vars(eval(var), kwarg_name, result_dict)

        initialized_cls = cls(**kwargs)
        for k, v in result_dict.items():
            setattr(initialized_cls, k, v)

        return initialized_cls

    def create_connection(self, return_client=True):
        http_auth = self._get_http_auth()
        client = connections.create_connection(
            hosts=self.hosts,
            http_auth=http_auth,
            timeout=self.timeout,
            use_ssl=self.use_ssl,
            verify_certs=self.verify_certs,
            ca_certs=self.ca_certs,
            client_cert=self.client_cert,
            client_key=self.client_key,
            headers=self.headers,
            connection_class=RequestsHttpConnection
        )

        if return_client:
            return client
        return



from __future__ import print_function

import os
import pytest
import sys
from flask import Flask
from mock import patch

PY3 = sys.version_info[0] >= 3

__all__ = ["app", "db", "session", "Person", "rhodb"]


@pytest.fixture(scope='function')
def app(request):
    app = Flask(__name__)
    app.config['DEBUG'] = True
    app.config['TESTING'] = True
    if hasattr(request.cls, 'database_name'):
        app.config['SQLALCHEMY_ECHO'] = False
        app.config['SQLALCHEMY_DATABASE_URI'] =\
            'postgresql://localhost/{0}'.format(request.cls.database_name)

    if hasattr(request.cls, 'app_config'):
        app.config.update(request.cls.app_config)

    ctx = app.app_context()
    ctx.push()

    def teardown():
        ctx.pop()

    request.addfinalizer(teardown)

    return app


@pytest.fixture(scope='function')
def db(app, request):
    """Session-wide test database."""
    scenario = 'with-app'
    if 'scenario' in request.funcargnames:
        scenario = request.getfuncargvalue('scenario')

    if "rhodb.database" in sys.modules:
        del sys.modules['rhodb.database']

    if "rhodb.flask_sqlalchemy" in sys.modules:
        del sys.modules['rhodb.flask_sqlalchemy']

    if "rhodb.rhodb_sqlalchemy" in sys.modules:
        del sys.modules['rhodb.rhodb_sqlalchemy']

    if 'flask_sqlalchemy' in sys.modules:
        del sys.modules['flask_sqlalchemy']

    # Check if we're running a parametrized test

    # If the parametrized test is without-app, then set the database connection
    # uri
    if scenario == 'without-app':
        os.environ['SQLALCHEMY_DATABASE_URI'] =\
            'postgresql://localhost/{0}'.format(request.cls.database_name)
        os.environ['SQLALCHEMY_ECHO'] = 'False'
        os.environ['SQLALCHEMY_POOL_SIZE'] = '5'
        os.environ['SQLALCHEMY_POOL_TIMEOUT'] = '30'
        os.environ['SQLALCHEMY_POOL_RECYCLE'] = '-1'
        os.environ['SQLALCHEMY_MAX_OVERFLOW'] = '10'

    # Mock the import statement so that we that the db variable in
    # rhodb.database can be set to the proper SQLAlchemy instance
    orig_import = __import__

    def import_mock(name, *args, **kwargs):
        if name == 'flask_sqlalchemy' and scenario == 'without-app':
            raise ImportError('package not found')

        return orig_import(name, *args, **kwargs)

    import_obj = 'builtins.__import__' if PY3 else '__builtin__.__import__'
    with patch(import_obj, side_effect=import_mock):
        from rhodb.database import SQLAlchemy, create_database, drop_database
        _db =  SQLAlchemy()

        try:
            drop_database(request.cls.database_name)
            create_database(request.cls.database_name)
        except:
            pass

        if scenario == 'with-app':
            _db.init_app(app)
        if scenario == 'without-app':
            _db.init_session()

        def teardown():
            _db.session.rollback()
            _db.session.close()
            _db.session.bind.dispose()

        request.addfinalizer(teardown)
        return _db


@pytest.fixture(scope='function')
def session(db, request):
    """Creates a new database session for a test."""

    db.session.begin_nested()

    def teardown():
        db.session.rollback()
        db.session.close()
        db.session.bind.dispose()

    request.addfinalizer(teardown)
    return db.session


@pytest.fixture(scope='function',
                params=['postgres', 'sqlite', 'sqlite-mem', 'mysql'])
def rhodb(request):
    from rhodb.database import create_database, drop_database
    sqlite_db = '/tmp/{0}.db'.format(request.cls.database_name)

    try:
        if request.param in ['postgres', 'mysql']:
            drop_database(request.cls.database_name, request.param)
            create_database(request.cls.database_name, request.param)
        elif request.param == 'sqlite':
            os.remove(sqlite_db)

    except Exception as e:
        print(e)
        pass

    from rhodb.rhodb_sqlalchemy import SQLAlchemy

    os.environ['SQLALCHEMY_ECHO'] = 'False'
    if 'SQLALCHEMY_POOL_SIZE' in os.environ:
        del os.environ['SQLALCHEMY_POOL_SIZE']
    if 'SQLALCHEMY_POOL_TIMEOUT' in os.environ:
        del os.environ['SQLALCHEMY_POOL_TIMEOUT']
    if 'SQLALCHEMY_POOL_RECYCLE' in os.environ:
        del os.environ['SQLALCHEMY_POOL_RECYCLE']
    if 'SQLALCHEMY_MAX_OVERFLOW' in os.environ:
        del os.environ['SQLALCHEMY_MAX_OVERFLOW']

    if request.param == 'postgres':
        os.environ['SQLALCHEMY_DATABASE_URI'] =\
            'postgresql://localhost/{0}'.format(request.cls.database_name)
    elif request.param == 'sqlite':
        os.environ['SQLALCHEMY_DATABASE_URI'] =\
            'sqlite:///{0}'.format(sqlite_db)
    elif request.param == 'sqlite-mem':
        os.environ['SQLALCHEMY_DATABASE_URI'] = 'sqlite://'
    elif request.param == 'mysql':
        os.environ['SQLALCHEMY_DATABASE_URI'] =\
            'mysql://root@localhost/{0}'.format(request.cls.database_name)
    else:
        raise Exception('Driver {0} not supported'.format(request.param))

    _db = SQLAlchemy()
    _db.init_session()

    def teardown():
        _db.session.rollback()
        _db.session.close()
        _db.session.bind.dispose()

    request.addfinalizer(teardown)
    return _db


@pytest.fixture(scope='function')
def Person(db, request):
    scenario = 'with-app'
    if 'scenario' in request.funcargnames:
        scenario = request.getfuncargvalue('scenario')

    orig_import = __import__

    def import_mock(name, *args, **kwargs):
        if name == 'flask_sqlalchemy' and scenario == 'without-app':
            raise ImportError('package not found')

        return orig_import(name, *args, **kwargs)

    import_obj = 'builtins.__import__' if PY3 else '__builtin__.__import__'
    with patch(import_obj, side_effect=import_mock):
        from rhodb.database import BaseWithIntPK, MutableDict, JSONEncodedDict

        class Person(BaseWithIntPK, db.Model):
            __tablename__ = 'person'

            given_name = db.Column(db.String(255))
            family_name = db.Column(db.String(255))
            custom_data = db.Column(MutableDict.as_mutable(JSONEncodedDict),
                                    default={})
            birth_date = db.Column(db.Interval)

            __serializable_columns__ = ['id', 'given_name', 'family_name',
                                        'birth_date']

            def __init__(self, given_name='Ricky', family_name='Bobby',
                         custom_data={'age': 30}):
                self.given_name = given_name
                self.family_name = family_name
                self.custom_data = custom_data

        db.create_all()

        return Person



"""
TODO: DEPRECATE. Future apps should simply use boto/3 directly, this doesn't add
anything.
"""

import logging
from boto3.session import Session

logger = logging.getLogger(__name__)


class S3Connector(object):
    """ Methods to interact with an S3 bucket

    Args:
        aws_access_key_id: The AWS ACCESS KEY ID
        aws_secret_access_key: The AWS SECRET ACCESS KEY
        aws_region: The AWS region. Defaults to `us-east-1`
    """
    def __init__(self, aws_access_key_id, aws_secret_access_key,
                 aws_region='us-east-1'):
        session = Session(aws_access_key_id=aws_access_key_id,
                          aws_secret_access_key=aws_secret_access_key,
                          region_name=aws_region)
        self.client = session.client('s3')

    def list_objects(self, bucket, folder=''):
        """ List the objects in a specific bucket and optionally folder.

            This strips out the metadata that is sent along with the
            list_objects() method and returns only the Contents list.

        Args:
            bucket: The S3 bucket
            folder: A folder within a bucket. Defaults to empty string.
        """
        try:
            retval = self.client.list_objects(
                Bucket=bucket,
                Delimiter='/',
                Prefix=folder
            )
        except Exception as e:
            logger.exception(e)
            return None

        if 'Contents' not in retval.keys():
            return None

        return retval['Contents']

    def put(self, bucket, key, body, metadata={}):
        """ Put an object into a specific bucket with a key

        Args:
            bucket: The S3 bucket
            key: The object's `key`
            body: The object's `body`
            metadata: Any metadata to associate with the object.
        """
        try:
            self.client.put_object(
                Bucket=bucket,
                Key=key,
                Body=body,
                Metadata=metadata
            )
            return True
        except Exception as e:
            logger.exception(e)
            return False

    def upload_file(self, bucket, key, file_path):
        """ Uploads a local file to a S3 bucket. It performs multipart uploads
        for large files.

        Args:
            bucket: The S3 bucket
            key: The object's `key`
            file_path: Location of the file to upload

        Retruns: True if the file was uploaded, False otherwise.
        """

        try:
            self.client.upload_file(
                Filename=file_path,
                Bucket=bucket,
                Key=key
            )
        except Exception as e:
            logger.exception(e)
            return False

        return True

    def get(self, bucket, key):
        """ Retrieve a specific key's value from a bucket

        Args:
            bucket: The S3 bucket
            key: The object's `key`
        """
        try:
            return self.client.get_object(
                Bucket=bucket,
                Key=key
            )
        except Exception as e:
            logger.exception(e)
            return None

    def get_meta(self, bucket, key):
        """ Retrieve a specific key's meta data from a bucket

        Args:
            bucket: The S3 bucket
            key: The object's `key`
        """
        try:
            return self.client.head_object(
                Bucket=bucket,
                Key=key
            )
        except Exception as e:
            logger.exception(e)
            return None

    def copy_object(self, src_bucket, src_key, metadata=None,
                    dest_bucket=None, dest_key=None):
        """ Copy an object. If no destination bucket/key provided, then update.

            Allows optional update to object's meta data.

        Args:
            src_bucket: The S3 source bucket from which to copy.
            src_key: The key of the object to copy.
            metadata: Any new metadata for the object.
            dest_bucket: The destination bucket. If None, the source bucket is
                used instead.
            dest_key: The key of the destination object. If None, the source
                key is used instead.
        """
        metadata_directive = 'COPY' if metadata is None else 'REPLACE'
        metadata = {} if metadata is None else metadata

        if dest_bucket is None:
            dest_bucket = src_bucket

        if dest_key is None:
            dest_key = src_key

        try:
            return self.client.copy_object(
                Bucket=dest_bucket,
                Key=dest_key,
                Metadata=metadata,
                MetadataDirective=metadata_directive,
                CopySource=src_bucket + "/" + src_key
            )
        except Exception as e:
            logger.exception(e)
            return None

    def delete(self, bucket, key):
        """ Delete a specific key

        Args:
            bucket: The S3 bucket
            key: The object's `key`
        """
        try:
            return self.client.delete_object(
                Bucket=bucket,
                Key=key
            )
        except Exception as e:
            logger.exception(e)
            return None

    def delete_multiple(self, bucket, keys):
        """ Delete a set of keys. Expects an array of objects with 'Key' key

        Args:
            bucket: The S3 bucket
            keys: The list of keys
        """
        try:
            return self.client.delete_objects(
                Bucket=bucket,
                Delete={
                    'Objects': keys
                }
            )
        except Exception as e:
            logger.exception(e)
            return None



from __future__ import print_function

import os
import redis
import logging

from collections import namedtuple
from redis import __version__ as redis_version
from redis.sentinel import Sentinel

from .compat import urlparse, iteritems

logger = logging.getLogger(__name__)


SentinelConfig =\
    namedtuple('SentinelConfig', ['hosts', 'options', 'service_name'])


class FlaskRedis(object):

    EXTENSION_NAME = 'flask_redis'

    def __init__(self, app=None, using_sentinel=False, **kwargs):
        if app is not None:
            self.init_app(app, using_sentinel, **kwargs)

    def init_app(self, app, using_sentinel=False, **kwargs):
        redis_url = app.config.get('REDIS_URL', None)

        if not hasattr(app, 'extensions'):
            app.extensions = {}

        if using_sentinel:
            if redis_url is None:
                self.master = None
                self.slave = None
            else:
                sentinel = SentinelRedisConnector(redis_url)
                self.master = sentinel.get_master_connection()
                self.slave = sentinel.get_slave_connection()
        else:
            if redis_url is None:
                self.master = None
            else:
                rc = RedisConnector(redis_url, **kwargs)
                self.master = rc.get_connection()
        app.extensions[self.EXTENSION_NAME] = self
        self.app = app


class SentinelRedisConnector(object):

    def __init__(self, sentinel_url):
        self.sentinel_config = self._parse_url(sentinel_url)
        options = self.sentinel_config.options
        if (redis_version == '2.9.1'):
            options = {
                'password': self.sentinel_config.options.get('password', None),
                'socket_timeout':
                    self.sentinel_config.options.get('socket_timeout', None),
                'min_other_sentinels':
                    self.sentinel_config.options.get('min_other_sentinels', 0),
            }

        self.sentinel = Sentinel(self.sentinel_config.hosts, **options)

    def _parse_url(self, sentinel_url):
        if sentinel_url is None:
            raise ValueError('Invalid conneciton url: {}'.format(sentinel_url))

        url = urlparse.urlparse(sentinel_url)
        if url.scheme != 'redis+sentinel':
            raise ValueError('Unsupported scheme: {}'.format(url.scheme))

        path_parts = url.path.split('/')
        if len(path_parts) < 3:
            raise ValueError("Invalid path: {0}. Must be of the form "
                             "/cluster-name/db-number".format(url.path))

        sentinel_option_types = {
            'min_other_sentinels': int,
            'socket_timeout': float
        }
        connection_options = {
            'db': int(path_parts[2])
        }

        if '@' in url.netloc:
            auth, hostspec = url.netloc.split('@', 1)

            if ':' in auth:
                _, password = auth.split(':', 1)
                connection_options['password'] = password
        else:
            hostspec = url.netloc

        hosts = self._parse_hosts(hostspec)

        for name, value in iteritems(urlparse.parse_qs(url.query)):
            if name not in sentinel_option_types:
                continue

            if len(value) > 1:
                raise ValueError('Multiple values specified for {}'
                                 .format(name))

            _type = sentinel_option_types[name]
            connection_options[name] = _type(value[0])

        return SentinelConfig(hosts, connection_options, path_parts[1])

    def _parse_hosts(self, hostspec):
        hosts = []
        for hp in hostspec.split(','):
            if ':' in hp:
                host, port = hp.split(':', 1)
                port = int(port)
            else:
                host = hp
                port = 26379  # default sentinel port

            hosts.append((host, port))

        return hosts

    def get_master_connection(self):
        return self.sentinel.master_for(self.sentinel_config.service_name,
                                        **self.sentinel_config.options)

    def get_slave_connection(self):
        return self.sentinel.slave_for(self.sentinel_config.service_name,
                                       **self.sentinel_config.options)


class RedisConnector(object):

    def __init__(self, redis_url, max_connections=None, **connection_options):
        url_bits = urlparse.urlparse(redis_url)
        options = {
            'host': url_bits.hostname,
            'port': int(url_bits.port),
            'db': url_bits.path.strip('/'),
            'password': url_bits.password
        }
        options.update(connection_options)
        self.connection_options = options

        # In v2.9.1 the StrictRedis does not respect max_connections when
        # creating it's connection pool so we create it manually first.
        if max_connections is not None and redis_version == '2.9.1':
            connection_pool = self._create_connection_pool(max_connections)
            self.connection_options['connection_pool'] = connection_pool

    def _create_connection_pool(self, max_connections):
            options = {
                'max_connections': max_connections
            }
            options.update(self.connection_options)
            return redis.ConnectionPool(**options)

    def get_connection(self):
        return redis.StrictRedis(**self.connection_options)



import datetime


def stringify_dates(x):
    """ Method that returns a string representation of a date.
        Useful for JSON serialization.
    """

    if isinstance(x, (datetime.datetime, datetime.date, datetime.time)):
        return x.isoformat()
    elif isinstance(x, datetime.timedelta):
        return x.total_seconds()
    return x



from copy import copy

class MockS3Client(object):

    _objects = {}

    def list_objects(self, Bucket, Delimiter=None, Prefix=None):
        if not Prefix:
            return {'Contents': self._objects[Bucket]}

        objs = []
        for key, obj in self._objects[Bucket].iteritems():
            if key.startsWith(Prefix):
                objs.append(obj)
        return {'Contents': objs}

    def put_object(self, Bucket, Key, Body, Metadata=None):
        if Bucket not in self._objects:
            self._objects[Bucket] = {}

        self._objects[Bucket][Key] = {
            'body': Body,
            'metadata': Metadata
        }

    def upload_file(self, Filename, Bucket, Key, ExtraArgs=None, Callback=None,
                    Config=None):
        # Filename is a StringIO instance. Easier for testing than actually
        # creating a file.
        self.put_object(Bucket, Key, Body=Filename.getvalue())

    def get_object(self, Bucket, Key):
        return self._objects[Bucket][Key]

    def delete_object(self, Bucket, Key):
        obj = self.get_object(Bucket, Key)
        del self._objects[Bucket][Key]
        return obj

    def delete_objects(self, Bucket, Delete):
        objs = []
        for key in Delete['Objects']:
            objs.append(self.delete_object(Bucket, key))

        return objs

    def head_object(self, Bucket, Key):
        return self.get_object(Bucket, Key)['metadata']

    def copy_object(self, Bucket, Key, CopySource, Metadata=None,
                    MetadataDirective='COPY'):
        bucket, key = CopySource.split('/')
        obj = self.get_object(bucket, key)
        new_obj = copy(obj)
        if Metadata:
            new_obj['metadata'] = Metadata

        self.put_object(Bucket, Key, Body=new_obj['body'],
                        Metadata=new_obj['metadata'])
        return {}


class MockBotoSession(object):
    def __init__(self, *args, **kwargs):
        pass

    def client(self, service_name):
        return MockS3Client()



from arango import ArangoClient

try:
    from flask import _app_ctx_stack as stack
except ImportError:
    try:
        from flask import _request_ctx_stack as stack
    except:
        pass


class FlaskArangoDB(object):
    def __init__(self, app=None):
        self.app = app
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        self.app = app

    def connect(self):
        self.arango = ArangoDB(
            protocol='http',
            host=self.app.config['ARANGO_HOST'],
            port=self.app.config['ARANGO_PORT'],
            db_name=self.app.config['ARANGO_DB'],
            username=self.app.config['ARANGO_USERNAME'],
            password=self.app.config['ARANGO_PASSWORD']
        )
        return self.arango.client

    @property
    def client(self):
        ctx = stack.top
        if ctx is not None:
            if not hasattr(ctx, 'arango_client'):
                ctx.arango_client = self.connect()
            return ctx.arango_client

    @property
    def db(self):
        if self.client:
            db_name = self.app.config['ARANGO_DB']
            return self.client.db(name=db_name)


class ArangoDB(object):

    def __init__(self, protocol, host, port, db_name, username=None,
                 password=None):
        self.host = host
        self.port = port
        self.db_name = db_name
        self.username = username
        self.password = password
        self._client = None

    @property
    def client(self):
        if self._client is None:
            self._client = ArangoClient(
                protocol='http',
                host=self.host,
                port=self.port,
                username=self.username,
                password=self.password,
                enable_logging=True
            )
        return self._client

    @property
    def db(self):
        if self.client:
            return self.client.db(name=self.db_name)



import datetime
import json
import logging
import os
import ssl
import uuid

from cassandra import ConsistencyLevel
from cassandra.auth import PlainTextAuthProvider
from cassandra.cqlengine import connection, ValidationError, columns
from cassandra.cqlengine.models import Model
from cassandra.policies import DCAwareRoundRobinPolicy, TokenAwarePolicy

from .utils import stringify_dates

logger = logging.getLogger(__name__)


FLASK_EXTENSION_NAME = 'cassandra'


class CassandraDBCOnfig(object):

    hosts = None
    keyspace = None
    username = None
    password = None
    consistency_level = 'ALL'
    ssl_options = None


class FlaskCassandra(object):

    def __init__(self, app=None):
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        ssl_options = None
        if app.config.get('CASSANDRA_ENABLE_SSL', False):
            ssl_options = {
                'ssl_version': ssl.PROTOCOL_TLSv1,
                'ca_certs': app.config.get('CASSANDRA_SSL_CERT', None),
                'keyfile': app.config.get('CASSANDRA_USER_KEY', None),
                'certfile': app.config.get('CASSANDRA_USER_CERT', None)
            }

        cs = Cassandra(
            hosts=app.config.get('CASSANDRA_HOSTS', None),
            port=app.config.get('CASSANDRA_PORT', 9042),
            keyspace=app.config.get('CASSANDRA_KEYSPACE', None),
            username=app.config.get('CASSANDRA_USERNAME', None),
            password=app.config.get('CASSANDRA_PASSWORD', None),
            consistency_level=\
                app.config.get('CASSANDRA_CONSISTENCY_LEVEL', 'QUORUM'),
            ssl_options=ssl_options,
            local_region=app.config.get('CASSANDRA_LOCAL_REGION', None),
            remote_hosts=\
                app.config.get('CASSANDRA_REMOTE_HOSTS', 0)
        )

        session = cs.connect()

        # Store the extension in the app
        if not hasattr(app, 'extensions'):
            app.extensions = {}
        app.extensions[FLASK_EXTENSION_NAME] = cs
        self.app = app


class Cassandra(object):

    def __init__(self, hosts, port=9042, keyspace=None, username=None,
                 password=None, consistency_level=None, ssl_options=None,
                 local_region=None, remote_hosts=0):

        self.hosts = hosts
        self.port = port
        self.keyspace = keyspace
        self.username = username
        self.password = password
        self.consistency_level = consistency_level
        self.ssl_options = ssl_options
        self.local_region = local_region
        self.session = None

        # number of hosts in remote dc to connect to, default 0
        self.remote_hosts = remote_hosts

    def connect(self):
        # The cassandra uwsgi integration states that we should close any
        # existing connections before we open a new one. See:
        # https://datastax.github.io/python-driver/cqlengine/third_party.html#uwsgi
        # So, we try to close any existing connection first.
        self._disconnect()

        # Setup a session if is not already available
        if connection.session is None:
            kwargs = {}
            if self.username and self.password:
                kwargs['auth_provider'] = PlainTextAuthProvider(
                    username=self.username,
                    password=self.password
                )

            if self.ssl_options:
                kwargs['ssl_options'] = self.ssl_options

            if self.local_region:
                kwargs['load_balancing_policy'] = TokenAwarePolicy(
                    DCAwareRoundRobinPolicy(
                        self.local_region,
                        used_hosts_per_remote_dc=self.remote_hosts)
                )

            consistency = ConsistencyLevel.name_to_value.get(
                self.consistency_level, ConsistencyLevel.ALL)
            connection.setup(self.hosts, self.keyspace, consistency, **kwargs)
            logger.info("Opened connection to Cassandra cluster.")

            self.session = connection.session
        return connection.session

    def disconnect(self):
        self._disconnect()
        logger.info("Closed connection to Cassandra cluster.")

    def _disconnect(self):
        if connection.cluster is not None:
            connection.cluster.shutdown()
        if connection.session is not None:
            connection.session.shutdown()


class CassandraModel(Model):
    __abstract__ = True

    def to_dict(self):
        """ Converts a cassandra model to a dict.

        If `__serializable_columns__` is defined, then it uses the list of
        columns specified in this variable to create the dict.

        If `__unserializable_columns__`, it will use the list of defined columns
        from the model minus the list in `__unserializable_columns__` to
        create the dict.

        If neither `__serializable_columns__` nor `__unserializable_columns__`
        is specified, the it simply uses the list of defined columns in the
        model to create the dict.
        """
        if hasattr(self, '__serializable_columns__'):
            cols = self.__serializable_columns__
        elif hasattr(self, '__unserializable_columns__'):
            cols = [col for col in self._defined_columns.keys()
                    if col not in self.__unserializable_columns__]
        else:
            cols = self._defined_columns.keys()

        def _getattr(k):
            v = getattr(self, k)
            if isinstance(v, uuid.UUID):
                return str(v)
            else:
                return stringify_dates(v)

        retval = {k: _getattr(k) for k in cols}
        return retval

    def validate(self):
        # Some of the messages in the Cassandra validation errors are UGLY, so
        # let's replace them here
        try:
            super(CassandraModel, self).validate()
        except ValidationError as e:
            if 'None values are not allowed' in e.message:
                raise ValidationError(
                    e.message.replace('- None values are not allowed',
                                      'is required'))
            elif 'is shorter than 1 characters' in e.message:
                raise ValidationError(
                    e.message.replace('shorter than 1 characters', 'required'))
            else:
                raise

    @classmethod
    def get_by_pk(cls, pks):
        """ Get objects by primary key.

        Args:
            pks: Dict mapping pk field to list of values. Example:
                {
                    field_1: [val1, val2],
                    field_2: [val3, val4]
                }
        """

        q = cls.objects()
        for key, values in pks.iteritems():
            field = getattr(cls, key)
            q = q.filter(field.in_(values))

        return q.all()


class TextJSON(columns.Column):
    """
    Stores a UTF-8 encoded string
    """
    db_type = 'text'

    def to_python(self, value):
        try:
            if type(value) == dict:
                return value
            return json.loads(value)
        except ValueError:
            raise ValidationError("{0} {1} is not valid JSON"
                                  .format(self.column_name, value))

    def to_database(self, value):
        """
        Converts python value into database value
        """
        val = value
        if val is None and self.has_default:
            val = self.get_default()
        if type(val) != dict:
            raise ValidationError("{0} {1} can't be converted to JSON value"
                                  .format(self.column_name, value))
        return json.dumps(val)



import os
import re
import sqlalchemy
from sqlalchemy import create_engine, event
from sqlalchemy.engine.url import make_url
from sqlalchemy.ext.declarative import declarative_base, DeclarativeMeta
from sqlalchemy.orm import scoped_session, sessionmaker, class_mapper, Query
from sqlalchemy.orm.exc import UnmappedClassError

from .compat import itervalues

_camelcase_re = re.compile(r'([A-Z]+)(?=[a-z0-9])')


def _make_table(db):
    def _make_table(*args, **kwargs):
        if len(args) > 1 and isinstance(args[1], db.Column):
            args = (args[0], db.metadata) + args[1:]
        info = kwargs.pop('info', None) or {}
        info.setdefault('bind_key', None)
        kwargs['info'] = info
        return sqlalchemy.Table(*args, **kwargs)
    return _make_table


def _include_sqlalchemy(obj):
    for module in sqlalchemy, sqlalchemy.orm:
        for key in module.__all__:
            if not hasattr(obj, key):
                setattr(obj, key, getattr(module, key))
    # Note: obj.Table does not attempt to be a SQLAlchemy Table class.
    obj.Table = _make_table(obj)
    obj.event = event


def _should_set_tablename(bases, d):
    """Check what values are set by a class and its bases to determine if a
    tablename should be automatically generated.
    The class and its bases are checked in order of precedence: the class
    itself then each base in the order they were given at class definition.
    Abstract classes do not generate a tablename, although they may have set
    or inherited a tablename elsewhere.
    If a class defines a tablename or table, a new one will not be generated.
    Otherwise, if the class defines a primary key, a new name will be generated.
    This supports:
    * Joined table inheritance without explicitly naming sub-models.
    * Single table inheritance.
    * Inheriting from mixins or abstract models.
    :param bases: base classes of new class
    :param d: new class dict
    :return: True if tablename should be set
    """

    if '__tablename__' in d or '__table__' in d or '__abstract__' in d:
        return False

    if any(v.primary_key for v in itervalues(d) if isinstance(v, sqlalchemy.Column)):
        return True

    for base in bases:
        if hasattr(base, '__tablename__') or hasattr(base, '__table__'):
            return False

        for name in dir(base):
            attr = getattr(base, name)

            if isinstance(attr, sqlalchemy.Column) and attr.primary_key:
                return True


class _BoundDeclarativeMeta(DeclarativeMeta):

    def __new__(cls, name, bases, d):
        if _should_set_tablename(bases, d):
            def _join(match):
                word = match.group()
                if len(word) > 1:
                    return ('_%s_%s' % (word[:-1], word[-1])).lower()
                return '_' + word.lower()
            d['__tablename__'] = _camelcase_re.sub(_join, name).lstrip('_')

        return DeclarativeMeta.__new__(cls, name, bases, d)

    def __init__(self, name, bases, d):
        bind_key = d.pop('__bind_key__', None)
        DeclarativeMeta.__init__(self, name, bases, d)
        if bind_key is not None:
            self.__table__.info['bind_key'] = bind_key


class _QueryProperty(object):

    def __init__(self, sa):
        self.sa = sa

    def __get__(self, obj, type):
        try:
            mapper = class_mapper(type)
            if mapper:
                return type.query_class(mapper, session=self.sa.session())
        except UnmappedClassError:  # pragma: no cover
            return None  # pragma: no cover


class Model(object):

    query_class = Query
    query = None


class SQLAlchemy(object):
    """ This class is used to integrate SQLAlchemy outside of Flask
    applications.

    :param use_native_unicode: True or False. Used to tell the underlying
                               driver if it should use native unicode.
    :param session_options: Any options to pass when creating a session through
                            :func:`sqlalchemy.orm.sessionmaker` and
                            :func:`sqlalchemy.orm.scoped_session`.
    :param metadata: Any metadata to pass to
                     :func:`sqlalchemy.ext.declarative_base`.
    """

    def __init__(self, use_native_unicode=True, session_options=None,
                 metadata=None, database_uri=None, engine_options=None):
        self.use_native_unicode = use_native_unicode
        self.Model = self.make_declarative_base(metadata=metadata)
        self.engine = None
        self.session = None
        self.init_session(database_uri, session_options, engine_options)
        _include_sqlalchemy(self)

    @property
    def metadata(self):
        """ Returns the metadata """
        return self.Model.metadata

    def init_session(self, database_uri=None, session_options=None,
                     engine_options=None):
        if self.session:
            self.session.remove()
        self.engine = self.create_engine(engine_options, database_uri)
        self.session = self.create_scoped_session(session_options)

    def create_engine(self, options=None, database_uri=None):
        """ Creates a new engine with the given options and returns it.

        :param options: Subset of options to be passed to ``create_engine``
        """

        db_uri = database_uri if database_uri is not None\
            else os.environ.get('SQLALCHEMY_DATABASE_URI', 'sqlite://')

        info = make_url(db_uri)
        if options is None:
            options = {}
        self.apply_engine_options(options)
        self.apply_driver_hacks(info, options)

        return create_engine(info, **options)

    def make_declarative_base(self, metadata=None):
        base = declarative_base(cls=Model, name='Model',
                                metadata=metadata,
                                metaclass=_BoundDeclarativeMeta)
        base.query = _QueryProperty(self)
        return base

    def create_scoped_session(self, options=None):
        """ Creates a new scoped session.

        :param options: Subset of options to be passed to ``sessionmaker`` and
                        ``scoped_session``.
        """
        if options is None:
            options = {}

        scopefunc = None
        if 'scopefunc' in options:
            scopefunc = options['scopefunc']
            del options['scopefunc']

        session_factory = sessionmaker(bind=self.engine, **options)
        Session = scoped_session(session_factory, scopefunc=scopefunc)
        return Session

    def apply_engine_options(self, options):
        def isint(val):
            try:
                int(val)
                return True
            except:
                return False

        def _setdefault(optionkey, envkey):
            value = os.environ.get(envkey, None)
            if value is not None:
                if value.lower() == 'true':
                    options[optionkey] = True
                elif value.lower() == 'false':
                    options[optionkey] = False
                elif isint(value):
                    options[optionkey] = int(value)
                else:
                    options[optionkey] = value

        _setdefault('echo', 'SQLALCHEMY_ECHO')
        _setdefault('pool_size', 'SQLALCHEMY_POOL_SIZE')
        _setdefault('pool_timeout', 'SQLALCHEMY_POOL_TIMEOUT')
        _setdefault('pool_recycle', 'SQLALCHEMY_POOL_RECYCLE')
        _setdefault('max_overflow', 'SQLALCHEMY_MAX_OVERFLOW')

    def apply_driver_hacks(self, info, options):
        """This method is called before engine creation and used to inject
        driver specific hacks into the options.

        :param info: The URI info.
        :param options: dictionary of keyword arguments that will then be used
        to call the :func:`sqlalchemy.create_engine` function.
        The default implementation provides some saner defaults for things
        like pool sizes for MySQL and sqlite.
        """
        # This method was copied directly from Flask-SQLAlchemy
        if info.drivername.startswith('mysql'):
            info.query.setdefault('charset', 'utf8')
            if info.drivername != 'mysql+gaerdbms':
                options.setdefault('pool_size', 10)
                options.setdefault('pool_recycle', 7200)
        elif info.drivername == 'sqlite':
            pool_size = options.get('pool_size')
            detected_in_memory = False
            # we go to memory and the pool size was explicitly set to 0
            # which is fail.  Let the user know that
            if info.database in (None, '', ':memory:'):
                detected_in_memory = True
                from sqlalchemy.pool import StaticPool
                options['poolclass'] = StaticPool
                if 'connect_args' not in options:
                    options['connect_args'] = {}
                options['connect_args']['check_same_thread'] = False

                if pool_size == 0:
                    raise RuntimeError('SQLite in memory database with an '
                                       'empty queue not possible due to data '
                                       'loss.')
            # if pool size is None or explicitly set to 0 we assume the
            # user did not want a queue for this sqlite connection and
            # hook in the null pool.
            elif not pool_size:
                from sqlalchemy.pool import NullPool
                options['poolclass'] = NullPool

            # if it's not an in memory database we make the path absolute.
            if not detected_in_memory:
                info.database = info.database

        if not self.use_native_unicode:
            options['use_native_unicode'] = False

    def create_all(self):
        self.metadata.create_all(self.engine)

    def drop_all(self):
        self.metadata.drop_all(self.engine)



#
# In this file, we're putting all database related
# code that multiple applications will need.  E.g.
# the db initialization, helper functions, and
# base classes for models.
#
import contextlib
import datetime
import json
import logging
import os
import sys
import uuid

from dateutil.parser import parse as dtparse

from itertools import chain

from sqlalchemy import Column, DateTime, Integer, String
from sqlalchemy.engine.url import make_url
from sqlalchemy.sql.expression import func
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.mutable import Mutable
from sqlalchemy.orm.properties import ColumnProperty
from sqlalchemy.types import TypeDecorator, TEXT

from .compat import str_types, iteritems
from .utils import stringify_dates

try:
    # This will throw an exception if Flask-SQLAlchemy is not in the
    # environment.
    from .flask_sqlalchemy import SQLAlchemy
except:
    # Flask-SQLAlchemy is not in the environment, then simply use our
    # own wrapper for SQLAlchemy
    from .rhodb_sqlalchemy import SQLAlchemy

logger = logging.getLogger(__name__)


class Base(object):
    added_datetime = Column(DateTime, default=func.now())
    changed_datetime = Column(
        DateTime,
        default=func.now(),
        onupdate=func.current_timestamp()
    )

    def update_columns(self, **kwargs):
        """ Allows you to overwrite values from an unpacked dictionary
            like cls.update_columns(foo=10, bar="kyle")
        """
        for field, field_value in iteritems(kwargs):
            if hasattr(self, field):
                setattr(self, field, field_value)

    @classmethod
    def pk_exists(cls, pks, pk='id'):
        """ Given a list of pimary keys, returns True or False, depending
            on if each pk exists in the database or not.
        """
        field = getattr(cls, pk)
        query = cls.query.with_entities(field).filter(field.in_(pks))
        found_set = set(result[0] for result in query)
        return [(pk in found_set) for pk in pks]

    @classmethod
    def get_by_pk(cls, pks, pk='id'):
        """ Get objects by primary key. Can return either a list or
            a single instance depending on if it is passed a list or not
            as input.
        """
        field = getattr(cls, pk)
        if isinstance(pks, list):
            result = cls.query.filter(field.in_(pks)).all()
        else:
            result = cls.query.filter(field == pks).one()
        return result

    @classmethod
    def _column_type(cls, kls, field):
        """ Returns the python type for a particular model column/field.
        """
        try:
            return getattr(kls, field).property.columns[0].type.python_type
        except NotImplementedError:
            logger.info("Couldn't get property type on field {0}.{1}"
                        .format(kls, field))
        except Exception as e:  # pragma: no cover
            logger.exception(e)  # pragma: no cover
            return None  # pragma: no cover

    @classmethod
    def _column_is_timedelta(cls, kls, field):
        if cls._column_type(kls, field) == datetime.timedelta:
            return True
        return False

    @staticmethod
    def _parse_datetime(datetime_string):
        """ Takes a string, returns a datetime.datetime instance
        """
        return dtparse(datetime_string)

    @classmethod
    def transform_datetime_fields(cls, d):
        """ Takes a dictionary and returns a copy of it in which
            datetime fields are appropriate python objects, instead
            of e.g. ints or floats or strings. Alters d in place.
        """
        for field, field_value in iteritems(d):
            if field_value is None:
                continue
            if 'datetime' in field:
                field_value = cls._parse_datetime(field_value)
            elif cls._column_is_timedelta(cls, field):
                field_value = datetime.timedelta(
                    seconds=field_value
                )
            d[field] = field_value

    @classmethod
    def from_dict(cls, d, transform_datetimes=True):
        """ Create an instance of the model from a dictionary. If
            `transform_datetimes` is True and a field looks like a
            datetime, we'll try to transform it as such. This is
            useful for
        """
        entity = cls()
        cls.transform_datetime_fields(d)
        for field, field_value in iteritems(d):
            if field_value is not None:
                try:
                    setattr(entity, field, field_value)
                except NotImplementedError:  # pragma: no cover
                    logger.info("Couldn't set attribute {0} on object {1} to "
                                "{2}".format(field, entity, field_value))   # pragma: no cover
                except Exception as e:  # pragma: no cover
                    logger.exception(e)  # pragma: no cover
        return entity

    @classmethod
    def equality_filter(cls, *args, **kwargs):
        """ Can be passed dictionaries and kwargs to
            do a filter on this class. So, you can, e.g. do
            Race.equality_filter(track_id='rir', series_id='W')
        """
        q = cls.query

        def f(q, x):
            for k, v in iteritems(x):
                q = q.filter(getattr(cls, k) == v)
            return q

        for arg in args:
            q = f(q, arg)
        q = f(q, kwargs)
        return q

    def duplicate(self, **overloads):
        """ Creates a copy of an instance of a model, excluding
            all columns that are primary keys. Will apply overloads
            after making the duplicate. Does not commit.
        """
        duplicate = self.__class__()
        primary_keys = set(self._primary_key_columns)
        for col in self._columns:
            if col not in primary_keys:
                setattr(duplicate, col, getattr(self, col))

        for col, value in iteritems(overloads):
            setattr(duplicate, col, value)

        return duplicate

    @declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()

    def __repr__(self):
        return '<{0} instance>'.format(self.__class__.__name__)

    @staticmethod
    def stringify_dates(x):
        return stringify_dates(x)

    @staticmethod
    def id_dammit(object_or_id):
        """ Get the id of an object
        """
        if hasattr(object_or_id, 'id'):
            return object_or_id.id
        return object_or_id

    @classmethod
    def instance_dammit(cls, object_or_id, pk='id'):
        """ Get an object
        """
        if isinstance(object_or_id, cls):
            return object_or_id
        return cls.query.filter(getattr(cls, pk) == object_or_id).one()

    @property
    def _primary_key_columns(self):
        """ Get a list of all the columns that are part of the
            primary key for this object.
        """
        return [
            a.key for a in
            self.__mapper__.primary_key
            if isinstance(a, Column)
        ]

    @property
    def _columns(self):
        """ Get all the columns that are defined for this model
        """
        return [
            a.key for a in
            self.__mapper__.attrs
            if isinstance(a, ColumnProperty)
        ]

    def to_dict(self, string_dates=True, extra_columns=None):
        """ Returns a representation of this instances as a
            regular python dictionary.  Generally used for
            serializing to JSON. Only serializes those fields
            referred to in the __serializable_columns__ attribute.
        """
        if hasattr(self, '__serializable_columns__'):
            cols = self.__serializable_columns__
        elif hasattr(self, '__unserializable_columns__'):
            cols = [
                col for col in self._columns
                if col not in self.__unserializable_columns__
            ]
        else:
            raise NotImplementedError

        if not extra_columns:
            extra_columns = []

        def getattr_plus(k):
            v = getattr(self, k)
            if isinstance(v, uuid.UUID):
                v = str(v)
            elif string_dates:
                v = self.stringify_dates(v)
            return v

        retval = dict(
            (k, getattr_plus(k))
            for k in chain(cols, extra_columns)
        )
        return retval


class BaseWithIntPK(Base):
    id = Column(Integer, primary_key=True)


# The following two classes are used to implement the
# recipe shown here:
# http://docs.sqlalchemy.org/en/rel_0_8/orm/extensions/mutable.html
# for mutable JSON/dictionary columns.
#
# Looks like Postgresql's JSON type is not yet supported by SQLA.
#
class JSONEncodedDict(TypeDecorator):
    """ Represents an immutable structure as a json-encoded string.
    """

    impl = TEXT

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


class MutableDict(Mutable, dict):
    @classmethod
    def coerce(cls, key, value):
        """Convert plain dictionaries to MutableDict.

        :param key: The dictionary key.
        :param value: The dictionary value.
        """

        if not isinstance(value, MutableDict):
            if isinstance(value, dict):
                return MutableDict(value)

            # this call will raise ValueError
            return Mutable.coerce(key, value)
        else:
            return value

    def __setitem__(self, key, value):
        "Detect dictionary set events and emit change events."

        dict.__setitem__(self, key, value)
        self.changed()

    def __delitem__(self, key):
        "Detect dictionary del events and emit change events."

        dict.__delitem__(self, key)
        self.changed()


def run_db_commands(cmds, database, dialect):
    if dialect == 'postgres':
        import psycopg2
        conn = psycopg2.connect(database=database)
        conn.autocommit = True
    elif dialect == 'mysql':
        import MySQLdb
        conn = MySQLdb.connect(host='localhost',
                               user='root',
                               db=database,
                               autocommit=True)
    else:
        raise Exception('Dialect {0} not supported.'.format(dialect))

    if isinstance(cmds, str_types):
        cmds = [cmds]

    cur = conn.cursor()
    for cmd in cmds:
        cur.execute(cmd)
    cur.close()
    conn.close()


def run_postgres_commands(cmds, database='postgres'):
    """ Run Postgres commands with autocommit

    :param cmds: List of commands.
    :param database: Defaults to postgres.
    """
    return run_db_commands(cmds, database, 'postgres')


def run_mysql_commands(cmds, database='mysql'):
    """ Run MySQL commands with autocommit.

    :param: cmds: List of commands.
    :param database: Defaults to mysql.
    """
    return run_db_commands(cmds, database, 'mysql')


def create_tables(db):
    """ Creates all tables in the database based on the SQLAlchemy models
    """
    db.create_all()


def drop_tables(db):
    """ Drops all the tables from the database
    """
    db.drop_all()


def reload_tables(db):
    """ Re-creates all tables in the database
    """
    drop_tables(db)
    create_tables(db)


def create_database(dbname, dialect='postgres'):
    """ Creates a Postgres database using psycopg2.

    :param dbname: The name of the database.
    """
    run_db_commands('CREATE DATABASE {0}'.format(dbname), dialect, dialect)


def enable_hstore_extension(dbname):
    """ Enables the hstore extension on the given database.

    :param dbname: The name of the database.
    """
    run_postgres_commands('CREATE EXTENSION IF NOT EXISTS hstore', dbname)


def disable_hstore_extension(dbname, cascade=False):
    """ Disables the hstore extension on the given database.

    :param dbname: The name of the database.
    :param cascade: Whether to auto drop objects that depend on HSTORE
    """

    sql = 'DROP EXTENSION IF EXISTS hstore'
    if cascade:
        sql += ' CASCADE'

    run_postgres_commands(sql, dbname)


def drop_database(dbname, dialect='postgres'):
    """ Drops a Postgres database using psycopg2.

    :param dbname: The name of the database.
    """
    run_db_commands('DROP DATABASE IF EXISTS {0}'.format(dbname),
                    dialect,
                    dialect)


def clear_all_tables(db):
    """ Clears all tables in the database """

    with contextlib.closing(db.engine.connect()) as con:
        trans = con.begin()
        for table in reversed(db.metadata.sorted_tables):
            con.execute(table.delete())
        trans.commit()



import re
import ast
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    user_options = [
        ('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import coverage
        import pytest

        if self.pytest_args and len(self.pytest_args) > 0:
            self.test_args.extend(self.pytest_args.strip().split(' '))
            self.test_args.append('tests/')

        cov = coverage.Coverage()
        cov.start()
        errno = pytest.main(self.test_args)
        cov.stop()
        cov.report()
        cov.html_report()
        print("Wrote coverage report to htmlcov directory")
        sys.exit(errno)


_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('rhodb/__init__.py', 'rb') as f:
    __version__ = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

setup(
    name='rhodb',
    version=__version__,
    description="Pit Rho Generic Database Helpers",
    long_description=open('README.md', 'r').read(),
    maintainer="Pit Rho Corporation",
    license="Commercial",
    url="https://bitbucket.org/pitrho/rhodb",
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    install_requires=[],
    extras_require={
        'sqlalchemy': [
            'SQLAlchemy>=0.9.2,<1',
            'python-dateutil==2.4.2'
        ],
        'flask_sqlalchemy': [
            'SQLAlchemy>=0.9.2,<1',
            'Flask>=0.10.1,<1.0',
            'Flask-SQLAlchemy>=2.1,<3.0',
            'python-dateutil==2.4.2'
        ],
        'redis': ['redis>=2.9.1'],
        'cassandra': ['cassandra-driver>=3.1.1,<4'],
        'cassandra_redis': [
            'cassandra-driver>=3.1.1,<4',
            'redis>=2.9.1'
        ],
        'aws': [
            'boto3>=1.2.3,<2'
        ],
        'elasticsearch': [
            'elasticsearch>=5.0.0,<6.0.0',
            'requests-aws4auth>=0.9,<1.0',
            'elasticsearch-dsl>=5.2,<6.0',
        ],
        'arangodb': [
            'python-arango >=3.12.1, < 4'
        ],
        'test': [
            'tox==2.6.0'
        ]
    },
    cmdclass={'test': PyTest}
)



__version__ = '3.10.0'
VERSION = tuple(map(int, __version__.split('.')))


# Set default logging handler to avoid "No handler found" warnings.
import logging
try:  # Python 2.7+
    from logging import NullHandler
except ImportError:  # pragma: no cover
    class NullHandler(logging.Handler):  # pragma: no cover
        def emit(self, record):  # pragma: no cover
            pass  # pragma: no cover

logging.getLogger(__name__).addHandler(NullHandler())
